/**
 * @file startup.S
 * @brief Initialization procedure for Cortex-M3 and Cortex-M4.
 */

#include "define/hdr_special_registers.h"

.text
.balign 2
.syntax unified
.thumb
.thumb_func
.global Reset_Handler

Reset_Handler:

 /*
  * If we use both process stack and main stack, we need to initialize stack
  * pointer for PSP and set bit in CONTROL register.
  */

/*
 * Currently we don't use process stack and thread mode. It is handled by
 * FreeRTOS. So in linker script PSP size is set to 0.
 */
#if __process_stack_size > 0
	ldr		r0, =__process_stack_end
	msr		PSP, r0
	movs	r0, #CONTROL_ALTERNATE_STACK
	msr		CONTROL, r0
	isb
#endif /* __process_stack_size > 0 */

/*
 * Przejscie do funkcji inicjalizacyjnej low_level_init_0 dzialajacej przed
 * inicjalizacja sekcji data i bss.
 */
 /*
  * Go to init function low_level_init_0 working before data and bss section
  * initialization.
  */
	ldr		r0, =low_level_init_0
	blx		r0

/*
 * Initialization of RAM section with zeros.
 */
	movs	r0, #0
	ldr		r1, =__ram_start
	ldr		r2, =__ram_end

1:	cmp		r1, r2
	itt		lo
	strlo	r0, [r1], #4
	blo		1b


/*
 * Initialization of AUX section with zeros.
 */
	movs	r0, #0
	ldr		r1, =__aux_ram_start
	ldr		r2, =__aux_ram_end

1:	cmp		r1, r2
	itt		lo
	strlo	r0, [r1], #4
	blo		1b

/*
 * Initialization of data section with initial values.
 */
	ldr		r1, =__data_init_start
    ldr		r2, =__data_start
    ldr		r3, =__data_end

1:	cmp		r2, r3
	ittt	lo
	ldrlo	r0, [r1], #4
	strlo	r0, [r2], #4
	blo		1b

/*
 * C++ constructors for global and static objects.
 */
#ifdef __USES_CXX
	ldr		r0, =__libc_init_array
	blx		r0
#endif

/*
 * Go to initialization function low_level_init_1 working after data and
 * bss section initialization.
 */
	ldr		r0, =low_level_init_1
	blx		r0

/*
 * Go to main program.
 */
	ldr		r0, =main
	blx		r0

/*
 * C++ destructors for global and static objects.
 */
#ifdef __USES_CXX
	ldr		r0, =__libc_fini_array
	blx		r0
#endif

/*
 * Endless loop at the end of the program (after leaving main function)
 */
	b		.

/*
+=============================================================================+
| __default_low_level_init() - replacement for undefined low_level_init_0()
| and/or low_level_init_1(). This function just returns.
+=============================================================================+
*/

/*
 * If low_level_init_0 or low_level_init_1 is not implemented, empty function
 * __default_low_level_init is invoked.
 */
.text
.balign 2
.syntax unified
.thumb
.thumb_func
.global __default_low_level_init

__default_low_level_init:
	bx		lr

.weak	low_level_init_0
.global	low_level_init_0
.set	low_level_init_0, __default_low_level_init

.weak	low_level_init_1
.global	low_level_init_1
.set	low_level_init_1, __default_low_level_init
